@@ -34,33 +34,33 @@
 */
 
 :- module(ssl,
-          [ load_certificate/2,         % +Stream, -Certificate
-            load_private_key/3,         % +Stream, +Password, -Key
-            load_public_key/2,          % +Stream, -Key
-            load_crl/2,                 % +Stream, -Crl
-            system_root_certificates/1, % -List
-            cert_accept_any/5,          % +SSL, +ProblemCertificate,
-                                        % +AllCertificates, +FirstCertificate,
-                                        % +Error
-            ssl_context/3,              % +Role, -Config, :Options
-            ssl_add_certificate_key/4,  % +Config, +Cert, +Key, -Config
-            ssl_set_sni_hook/3,         % +Config, +Goal, -Config
-            ssl_negotiate/5,            % +Config, +PlainRead, +PlainWrite,
-                                        %          -SSLRead,   -SSLWrite
-            ssl_peer_certificate/2,     % +Stream, -Certificate
-            ssl_session/2               % +Stream, -Session
+          [ load_certificate/2,           % +Stream, -Certificate
+            load_private_key/3,           % +Stream, +Password, -Key
+            load_public_key/2,            % +Stream, -Key
+            load_crl/2,                   % +Stream, -Crl
+            system_root_certificates/1,   % -List
+            cert_accept_any/5,            % +SSL, +ProblemCertificate,
+                                          % +AllCertificates, +FirstCertificate,
+                                          % +Error
+            ssl_context/3,                % +Role, -Config, :Options
+            ssl_add_certificate_key/4,    % +Config, +Cert, +Key, -Config
+            ssl_set_options/3,            % +Config0, -Config, +Options
+            ssl_negotiate/5,              % +Config, +PlainRead, +PlainWrite,
+                                          %          -SSLRead,   -SSLWrite
+            ssl_peer_certificate/2,       % +Stream, -Certificate
+            ssl_peer_certificate_chain/2, % +Stream, -Certificates
+            ssl_session/2,                % +Stream, -Session
+            ssl_secure_ciphers/1          % -Ciphers
           ]).
-:- use_module(library(socket)).
-:- use_module(library(error)).
 :- use_module(library(option)).
-:- use_module(library(debug)).
+:- use_module(library(settings)).
 :- use_module(library(crypto), []).     % force initialization of libcrypto
 
 :- use_foreign_library(foreign(ssl4pl)).
 
 :- meta_predicate
     ssl_context(+, -, :),
-    ssl_set_sni_hook(+, 3, -).
+    ssl_set_options(+, -, :).
 
 :- predicate_options(ssl_context/3, 3,
                      [ host(atom),
@@ -107,6 +107,10 @@ easily be used.
 @see library(socket), library(http/http_open), library(crypto)
 */
 
+:- setting(secure_ciphers, atom,
+           'EECDH+AESGCM:EDH+AESGCM:EECDH+AES256:EDH+AES256:EECDH+CHACHA20:EDH+CHACHA20',
+           "Default set of ciphers considered secure").
+
 %!  ssl_context(+Role, -SSL, :Options) is det.
 %
 %   Create an  SSL context.  The context  defines several properties
@@ -203,7 +207,7 @@ easily be used.
 %     implementation that accepts any certificate.
 %     * cipher_list(+Atom)
 %     Specify a cipher preference list (one or more cipher strings
-%     separated by colons, commas or spaces).
+%     separated by colons, commas or spaces). See ssl_secure_ciphers/1.
 %     * ecdh_curve(+Atom)
 %     Specify a curve for ECDHE ciphers. If this option is not
 %     specified, the OpenSSL default parameters are used.  With
@@ -226,14 +230,14 @@ easily be used.
 %     very rarely a concern.
 %     * min_protocol_version(+Atom)
 %     Set the _minimum_ protocol version that can be negotiated.
-%     Atom is one of `sslv3`, `tlsv1`, `tlsv1_1` and `tlsv1_2`.
-%     This option is available with OpenSSL 1.1.0 and later, and
-%     should be used instead of `disable_ssl_methods/1`.
+%     Atom is one of `sslv3`, `tlsv1`, `tlsv1_1`, `tlsv1_2` and
+%     `tlsv1_3`. This option is available with OpenSSL 1.1.0 and
+%     later, and should be used instead of `disable_ssl_methods/1`.
 %     * max_protocol_version(+Atom)
 %     Set the _maximum_ protocol version that can be negotiated.
-%     Atom is one of `sslv3`, `tlsv1`, `tlsv1_1` and `tlsv1_2`.
-%     This option is available with OpenSSL 1.1.0 and later, and
-%     should be used instead of `disable_ssl_methods/1`.
+%     Atom is one of `sslv3`, `tlsv1`, `tlsv1_1`, `tlsv1_2` and
+%     `tlsv1_3`. This option is available with OpenSSL 1.1.0 and
+%     later, and should be used instead of `disable_ssl_methods/1`.
 %     * disable_ssl_methods(+List)
 %     A list of methods to disable. Unsupported methods will be
 %     ignored. Methods include `sslv2`, `sslv3`, `sslv23`,
@@ -303,15 +307,21 @@ ssl_copy_context(SSL0, SSL) :-
     ssl_context(server, SSL, []),
     '_ssl_init_from_context'(SSL0, SSL).
 
-%!  ssl_set_sni_hook(+SSL0, :Goal, -SSL)
+%!  ssl_set_options(+SSL0, -SSL, +Options)
 %
-%   SSL is the same as SSL0, except  that the SNI hook of SSL is Goal.
-%   See  the   sni_hook(:Goal)  option   of  ssl_context/3   for  more
-%   information about this hook.
+%   SSL is the same as SSL0, except for the options specified in
+%   Options.  The following options are supported: close_notify/1,
+%   close_parent/1, host/1, peer_cert/1, ecdh_curve/1,
+%   min_protocol_version/1, max_protocol_version/1,
+%   disable_ssl_methods/1, sni_hook/1, cert_verify_hook/1. See
+%   ssl_context/3 for more information about these options. This
+%   predicate allows you to tweak existing SSL contexts, which can be
+%   useful in hooks when creating servers with the HTTP
+%   infrastructure.
 
-ssl_set_sni_hook(SSL0, Goal, SSL) :-
+ssl_set_options(SSL0, SSL, Options) :-
     ssl_copy_context(SSL0, SSL),
-    '_ssl_set_sni_hook'(SSL, Goal).
+    '_ssl_set_options'(SSL, Options).
 
 %!  ssl_negotiate(+SSL,
 %!                +PlainRead, +PlainWrite,
@@ -350,6 +360,11 @@ ssl_set_sni_hook(SSL0, Goal, SSL) :-
 %       memberchk('CN' = CommonName), Subject)
 %     ==
 
+%!  ssl_peer_certificate_chain(+Stream, -Certificates) is det.
+%
+%   Certificates  is the  certificate  chain provided  by the  peer,
+%   represented as a list of certificates.
+
 %!  ssl_session(+Stream, -Session) is det.
 %
 %   Retrieves (debugging) properties from the SSL context associated
@@ -381,12 +396,15 @@ ssl_set_sni_hook(SSL0, Goal, SSL) :-
 %!  load_certificate(+Stream, -Certificate) is det.
 %
 %   Loads a certificate from a PEM- or DER-encoded stream, returning
-%   a term which will unify with   the same certificate if presented
-%   in cert_verify_hook. A certificate  is   a  list  containing the
-%   following terms: issuer_name/1, hash/1,  signature/1, version/1,
-%   notbefore/1,  notafter/1,  serial/1,   subject/1    and   key/1.
-%   subject/1  and  issuer_name  are  both    lists   of  =/2  terms
-%   representing the name.
+%   a term which  will unify with the same  certificate if presented
+%   in  cert_verify_hook. A  certificate  is a  list containing  the
+%   following    terms:    issuer_name/1,    hash/1,    signature/1,
+%   signature_algorithm/1,   version/1,   notbefore/1,   notafter/1,
+%   serial/1, subject/1 and key/1.   subject/1 and issuer_name/1 are
+%   both lists  of =/2  terms representing  the name.   With OpenSSL
+%   1.0.2 and  greater, to_be_signed/1  is also  available, yielding
+%   the hexadecimal representation of the TBS (to-be-signed) portion
+%   of the certificate.
 %
 %   Note that the OpenSSL `CA.pl`  utility creates certificates that
 %   have a human readable textual representation in front of the PEM
@@ -437,25 +455,25 @@ ssl_set_sni_hook(SSL0, Goal, SSL) :-
 
 %!  load_private_key(+Stream, +Password, -PrivateKey) is det.
 %
-%   Load a private key PrivateKey  from   the  given  stream Stream,
-%   using Password to decrypt the key if  it is encrypted. Note that
-%   the  password  is  currently  only   supported  for  PEM  files.
+%   Load  a private  key PrivateKey  from the  given stream  Stream,
+%   using Password to decrypt the key  if it is encrypted. Note that
+%   the  password  is  currently   only  supported  for  PEM  files.
 %   DER-encoded keys which are password protected will not load. The
-%   key must be an RSA key. EC, DH   and DSA keys are not supported,
-%   and PrivateKey will be  bound  to   an  atom  (ec_key, dh_key or
-%   dsa_key) if you try and load   such  a key. Otherwise PrivateKey
-%   will be unified with private_key(KeyTerm)   where  KeyTerm is a
-%   rsa/8 term representing an RSA key.
+%   key must be an RSA or EC key. DH and DSA keys are not supported,
+%   and PrivateKey will  be bound to an atom (dh_key  or dsa_key) if
+%   you  try and  load such  a  key.  Otherwise  PrivateKey will  be
+%   unified with private_key(KeyTerm) where KeyTerm is an rsa/8 term
+%   representing an RSA key, or ec/3 for EC keys.
 
 %!  load_public_key(+Stream, -PublicKey) is det.
 %
-%   Load a public key  PublicKey  from   the  given  stream  Stream.
+%   Load  a  public key  PublicKey  from  the given  stream  Stream.
 %   Supports loading both DER- and PEM-encoded keys. The key must be
-%   an RSA key. EC, DH and DSA keys are not supported, and PublicKey
-%   will be bound to an atom (one   of ec_key, dh_key or dsa_key) if
-%   you try and load such a key. Otherwise PublicKey will be unified
-%   with  public_key(KeyTerm)  where  KeyTerm  is    an  rsa/8  term
-%   representing an RSA key.
+%   an  RSA or  EC  key. DH  and  DSA keys  are  not supported,  and
+%   PublicKey will  be bound to an  atom (dh_key or dsa_key)  if you
+%   try and  load such  a key. Otherwise  PublicKey will  be unified
+%   with  public_key(KeyTerm)   where  KeyTerm  is  an   rsa/8  term
+%   representing an RSA key, or ec/3 for EC keys.
 
 
 %!  cert_accept_any(+SSL,
@@ -476,6 +494,31 @@ cert_accept_any(_SSL,
                 _ProblemCertificate, _AllCertificates, _FirstCertificate,
                 _Error).
 
+%!  ssl_secure_ciphers(-Ciphers:atom) is det.
+%
+%   Ciphers is a  secure cipher preference list that can  be used in the
+%   cipher_list/1 option of ssl_context/3.
+%
+%   Secure ciphers must guarantee forward secrecy, and must mitigate all
+%   known critical attacks.  As of  2018, using these ciphers allows you
+%   to obtain grade A on  https://www.ssllabs.com. For A+, you must also
+%   enable HTTP Strict  Transport Security (HSTS) by  sending a suitable
+%   header field in replies.
+%
+%   Note that obsolete ciphers *must* be   disabled  to reliably prevent
+%   protocol downgrade attacks.
+%
+%   The Ciphers list is read from   the setting `ssl:secure_ciphers` and
+%   can be controlled using  set_setting/2   and  other  predicates from
+%   library(settings).
+%
+%   *BEWARE*: This list must be changed when attacks on these ciphers
+%             become known! Keep an eye on this setting and adapt it
+%             as necessary in the future.
+
+ssl_secure_ciphers(Cs) :-
+    setting(secure_ciphers, Cs).
+
 
                  /*******************************
                  *           MESSAGES           *
