@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2011-2016, University of Amsterdam
+    Copyright (c)  2011-2017, University of Amsterdam
                               VU University Amsterdam
     All rights reserved.
 
@@ -68,7 +68,8 @@ locking is required.
 #define O_LOCALE 1
 #include <wchar.h>
 #define NEEDS_SWINSOCK
-#include "SWI-Stream.h"
+#define PL_STREAM_IMPL
+#include "pl-stream.h"
 #define PL_ARITY_AS_SIZE
 #include "SWI-Prolog.h"
 #include "pl-utf8.h"
@@ -307,55 +308,48 @@ print_trace(void)
 #endif /*DEBUG_IO_LOCKS*/
 
 
-int
-Slock(IOSTREAM *s)
-{ SLOCK(s);
-
-  if ( s->erased )
+static int
+S__lock(IOSTREAM *s)
+{ if ( s->erased )
   { SUNLOCK(s);
     return -1;
   }
 
-#ifdef DEBUG_IO_LOCKS
-  if ( s->locks > 2 )
-  { printf("  Lock [%d]: %s: %d locks", PL_thread_self(), Sname(s), s->locks+1);
-    print_trace();
+  if ( s->locks == 0 )
+  { if ( (s->flags & (SIO_NBUF|SIO_OUTPUT)) == (SIO_NBUF|SIO_OUTPUT) &&
+	 S__setbuf(s, NULL, TMPBUFSIZE) == (size_t)-1 )
+    { SUNLOCK(s);
+      return -1;
+    }
   }
-#endif
 
-  s->references++;
-  if ( !s->locks++ )
-  { if ( (s->flags & (SIO_NBUF|SIO_OUTPUT)) == (SIO_NBUF|SIO_OUTPUT) )
-      return S__setbuf(s, NULL, TMPBUFSIZE) == (size_t)-1 ? -1 : 0;
-  }
+  s->locks++;
+  Sreference(s);
 
   return 0;
 }
 
+int
+Slock(IOSTREAM *s)
+{ SLOCK(s);
+
+  return S__lock(s);
+}
+
 
 int
 StryLock(IOSTREAM *s)
 { if ( !STRYLOCK(s) )
     return -1;
 
-  if ( s->erased )
-  { SUNLOCK(s);
-    return -1;
-  }
-
-  if ( !s->locks++ )
-  { if ( (s->flags & (SIO_NBUF|SIO_OUTPUT)) == (SIO_NBUF|SIO_OUTPUT) )
-      return S__setbuf(s, NULL, TMPBUFSIZE) == (size_t)-1 ? -1 : 0;
-  }
-  s->references++;
-
-  return 0;
+  return S__lock(s);
 }
 
 
 int
 Sunlock(IOSTREAM *s)
 { int rval = 0;
+  int unalloc;
 
 #ifdef DEBUG_IO_LOCKS
   if ( s->locks > 3 )
@@ -373,9 +367,10 @@ Sunlock(IOSTREAM *s)
   { assert(0);
   }
 
-  s->references--;
+  unalloc = (Sunreference(s) == 0 && s->erased);
+
   SUNLOCK(s);
-  if ( s->references == 0 && s->erased )
+  if ( unalloc )
     unallocStream(s);
 
   return rval;
@@ -499,6 +494,10 @@ S__flushbuf(IOSTREAM *s)
 { char *from, *to;
   ssize_t rc;
 
+  if ( s->magic != SIO_MAGIC )
+  { errno = EINVAL;
+    return -1;
+  }
   SLOCK(s);
   from = s->buffer;
   to   = s->bufp;
@@ -529,7 +528,7 @@ S__flushbuf(IOSTREAM *s)
 	  errno = EPLEXCEPTION;
 	} else
 	  goto retry;
-      } else if ( errno != EPLEXCEPTION )
+      } else if ( errno != EPLEXCEPTION && !(s->flags&SIO_NOERROR) )
 	S__seterror(s);
       rc = -1;
       goto out;
@@ -594,6 +593,11 @@ int
 S__fillbuf(IOSTREAM *s)
 { int c;
 
+  if ( s->magic != SIO_MAGIC )
+  { errno = EINVAL;
+    return -1;
+  }
+
   if ( s->flags & (SIO_FEOF|SIO_FERR) )	/* reading past eof */
   { if ( s->flags & SIO_FEOF2ERR )
       Sseterr(s, (SIO_FEOF2|SIO_FERR), NULL);
@@ -671,7 +675,8 @@ S__fillbuf(IOSTREAM *s)
 	return -1;
 #endif
       } else
-      { S__seterror(s);
+      { if ( !(s->flags & SIO_FERR) )	/* error already set */
+	  S__seterror(s);
 	return -1;
       }
     }
@@ -1889,7 +1894,8 @@ Stell(IOSTREAM *s)
 
 void
 unallocStream(IOSTREAM *s)
-{
+{ S__destroyed(s);
+
 #ifdef O_PLMT
   if ( s->mutex )
   { recursiveMutexDelete(s->mutex);
@@ -1928,6 +1934,7 @@ Sclose(IOSTREAM *s)
   if ( s->upstream )
   { Sseterr(s, SIO_FERR, "Locked by upstream filter");
     reportStreamError(s);
+    Sunlock(s);
     return -1;
   }
 
@@ -2079,6 +2086,7 @@ Svprintf(const char *fm, va_list args)
 #define NEXTCHR(s, c)				\
 	switch (enc)				\
 	{ case ENC_ANSI:			\
+	  case ENC_ISO_LATIN_1:			\
 	    c = *(s)++; c &= 0xff;		\
 	    break;				\
 	  case ENC_UTF8:			\
@@ -2103,29 +2111,6 @@ Svprintf(const char *fm, va_list args)
 #define A_LEFT	0			/* left-aligned field */
 #define A_RIGHT 1			/* right-aligned field */
 
-#ifdef __WINDOWS__
-/* This is a copy from src/pl-nt.c, copied here to avoid dependency
-   of pl-stream.c on SWI-Prolog internals.
-*/
-
-static int
-ms_snprintf(char *buffer, size_t count, const char *fmt, ...)
-{ va_list ap;
-  int ret;
-
-  va_start(ap, fmt);
-  ret = _vsnprintf(buffer, count-1, fmt, ap);
-  va_end(ap);
-
-  if ( ret < 0 || ret == count )
-  { ret = (int)count;
-    buffer[count-1] = '\0';
-  }
-
-  return ret;
-}
-#endif
-
 #define SNPRINTF3(fm, a1) \
 	{ size_t __r; \
 	  assert(fs == fbuf); \
@@ -2235,6 +2220,10 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	    islong = INT_SIZE_T;
 	    fm++;
 	    break;
+	  case 'L':			/* %Ls: ISO-Latin-1 string */
+	    enc = ENC_ISO_LATIN_1;
+	    fm++;
+	    break;
 	  case 'U':			/* %Us: UTF-8 string */
 	    enc = ENC_UTF8;
 	    fm++;
@@ -2832,7 +2821,7 @@ If parent is an input stream we have
 
 	--> parent --> filter --> application
 
-This filter is referenced, so it won't be freed, even if it is closed.
+The filter is referenced, so it won't be freed, even if it is closed.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
@@ -2847,15 +2836,21 @@ Sset_filter(IOSTREAM *parent, IOSTREAM *filter)
     { errno = EINVAL;
       return -1;
     }
-    filter->references++;
+    Sreference(filter);
+    Sreference(parent);
+    assert(parent->upstream==NULL && filter->downstream==NULL);
     parent->upstream = filter;
     filter->downstream = parent;
     filter->timeout = parent->timeout;
   } else				/* clear filter */
-  { if ( parent->upstream )
-    { if ( --parent->upstream->references == 0 && parent->upstream->erased )
-	unallocStream(parent->upstream);
+  { if ( (filter=parent->upstream) )
+    { assert(filter->downstream == parent);
+      filter->downstream = NULL;
       parent->upstream = NULL;
+      if ( Sunreference(filter) == 0 && filter->erased )
+	unallocStream(filter);
+      if ( Sunreference(parent) == 0 && parent->erased )
+	unallocStream(parent);
     }
   }
 
@@ -3027,6 +3022,9 @@ Snew(void *handle, int flags, IOFUNCTIONS *functions)
   { errno = ENOMEM;
     return NULL;
   }
+
+  S__created(s);
+
   memset((char *)s, 0, sizeof(IOSTREAM));
   s->magic         = SIO_MAGIC;
   s->lastc         = EOF;
@@ -3636,7 +3634,7 @@ Sopenmem(char **buffer, size_t *sizep, const char* mode)
     Mode is "r" or "w".  The mode "rF" calls PL_free(*buffer) at when
     closed.
 
-Note: Its is NOT allows to access   streams  created with this call from
+Note: Its is NOT allowed to access   streams created with this call from
 multiple threads. This is ok for all   usage inside Prolog itself (often
 through tellString()/toldString(). This call is   intented  to use write
 and other output predicates to create strings.
@@ -3844,10 +3842,7 @@ Sclose_buffer(IOSTREAM *s)
 #define STDIO(n, f) { NULL, NULL, NULL, NULL, \
 		      EOF, SIO_MAGIC, 0, f, {0, 0, 0}, NULL, \
 		      (void *)(n), &Sttyfunctions, \
-		      0, NULL, \
-		      (void (*)(void *))0, NULL, \
 		      -1, \
-		      0, \
 		      ENC_ISO_LATIN_1 \
 		    }
 
