@@ -1746,23 +1746,23 @@ ar_pow(Number n1, Number n2, Number r)
     }
 
   { GET_LD				/* estimate the size, see above */
-    size_t  op1_bytes;
-    int64_t r_bytes;
+    size_t  op1_bits;
+    int64_t r_bits;
 
     switch(n1->type)
     { case V_INTEGER:
-	op1_bytes = MSB64(n1->value.i)+7/8;
+	op1_bits = MSB64(n1->value.i);
         break;
       case V_MPZ:
-	op1_bytes = mpz_sizeinbase(n1->value.mpz, 256);
+	op1_bits = mpz_sizeinbase(n1->value.mpz, 2);
         break;
       default:
 	assert(0);
         fail;
     }
 
-    if ( !( mul64(op1_bytes, exp, &r_bytes) &&
-	    r_bytes < (int64_t)globalStackLimit()
+    if ( !( mul64(op1_bits, exp, &r_bits) &&
+	    r_bits/8 < (int64_t)globalStackLimit()
 	  ) )
       return int_too_big();
   }
