@@ -139,6 +139,7 @@ const debug_topic debug_topics[] =
   DEBUG_TOPIC(MSG_SHIFT_POINTER),
   DEBUG_TOPIC(MSG_SHIFT_FRAME),
   DEBUG_TOPIC(MSG_STACK_OVERFLOW),
+  DEBUG_TOPIC(MSG_UNWIND_EXCEPTION),
 
   DEBUG_TOPIC(MSG_ATTVAR_LINK),
   DEBUG_TOPIC(MSG_CALL_RESIDUE_VARS),
@@ -173,6 +174,8 @@ const debug_topic debug_topics[] =
   DEBUG_TOPIC(MSG_ZIP),
   DEBUG_TOPIC(MSG_ZIP_STATE),
 
+  DEBUG_TOPIC(MSG_MODULE_REF),
+
   DEBUG_TOPIC(MSG_TABLING_WORK),
 
   DEBUG_TOPIC(CHK_SECURE),
