@@ -670,6 +670,18 @@ freeClauseListRef(ClauseRef cref)
 }
 
 
+static void
+vfree_clause_list_ref(void *cref)
+{ freeClauseListRef(cref);
+}
+
+
+static void
+lingerClauseListRef(Definition def, ClauseRef cref)
+{ linger(&def->lingering, vfree_clause_list_ref, cref);
+}
+
+
 static void
 unallocClauseIndexTableEntries(ClauseIndex ci)
 { ClauseBucket cb;
@@ -1086,7 +1098,7 @@ gcClauseBucket(Definition def, ClauseBucket ch,
 	}
 
 	if ( is_list )
-	  freeClauseListRef(c);
+	  lingerClauseListRef(def, c);
 	else
 	  lingerClauseRef(c);
 
@@ -1846,6 +1858,8 @@ checkClauseIndexes(Definition def)
 
       if ( ISDEADCI(ci) )
 	continue;
+      if ( ci->incomplete )		/* building async; stats may be wrong */
+	continue;
 
       for(i=0,cb=ci->entries; i<ci->buckets; i++,cb++)
       { ClauseRef cref;
